--This addon is marked as reusable, so do what you want with it. ~Krabs

--Which skin gets the rebound dash?
local rbdskin = "sonic"

--Dash
local DASHLENGTH = 24
local DASHZPERCENT = 50

--Factors
local WATERFACTOR = FRACUNIT * 2/3
local SUPERFACTOR = FRACUNIT * 4/3

--Wall bounce vertical
local WALLBOUNCEZ = 12 * FRACUNIT
local WEAKWALLBOUNCEZ = 4 * FRACUNIT
local NOCLWALLBOUNCEZ = 3 * FRACUNIT
local SECONDBOUNCEZMULT = 120

--Wall bounce thrust based on direction of the wall
local WALLTHRUST = 5 * FRACUNIT
local WEAKWALLTHRUST = FRACUNIT / 5
local NOCLIMBWALLTHRUST = 3 * FRACUNIT

--Wall bounce reflected speed accross the wall
local BOUNCEPERCENT = 16
local WEAKBOUNCEPERCENT = 3

--Badnik bounce
local BADNIKZMULT = 120
local BADNIKSPEED = 80
local BADNIKBUMP = 20

--Assign sprites, objects, and sfx
freeslot(
	"SPR_REBO",
	"MT_REBOUND",
	"S_REBOUND",
	"sfx_airdsh",
	"sfx_bounc1",
	"sfx_bounc2"
)

--Closed captions
sfxinfo[sfx_airdsh].caption = "Dash"
sfxinfo[sfx_bounc1].caption = "Rebound"
sfxinfo[sfx_bounc2].caption = "Heavy Rebound"
sfxinfo[sfx_s3k82].caption = "/"

--Wallbounce ring effect object
mobjinfo[MT_REBOUND] = {
	doomednum = -1,
	spawnstate = S_REBOUND,
	flags = MF_NOCLIP|MF_SCENERY|MF_NOGRAVITY
}
states[S_REBOUND] = {
	sprite = SPR_REBO,
	frame = TR_TRANS50|FF_PAPERSPRITE|A
}

--FixedLerp
local function FixedLerp(a, b, w)
    return FixedMul((FRACUNIT - w), a) + FixedMul(w, b)
end

--Functions needed to do the bounce calculations (these could also be used to allow bounces off of sloped terrain, but I don't think we need to do that for this ability)
local function SphereToCartesian(alpha, beta)
	local t = {}

	t.x = FixedMul(cos(alpha), cos(beta))
	t.y = FixedMul(sin(alpha), cos(beta))
	t.z = sin(beta)

	return t
end

local function FixedDotProduct3D(a, b)
	return FixedMul(a.x, b.x) + FixedMul(a.y, b.y) + FixedMul(a.z, b.z)
end

local function FixedScalar3D(vect, scalar)
	local vect2 = {}
	vect2.x = FixedMul(vect.x, scalar)
	vect2.y = FixedMul(vect.y, scalar)
	--vect2.z = FixedMul(vect.z, scalar)
	vect2.z = vect.z
	
	return vect2
end

local function Percent3D(vect, percent)
	local vect2 = {}
	vect2.x = vect.x * percent / 100
	vect2.y = vect.y * percent / 100
	--vect2.z = FixedMul(vect.z, scalar)
	vect2.z = vect.z
	
	return vect2
end

local function VectAdd3D(vecta, vectb)
	local vsum = {}
	vsum.x = vecta.x + vectb.x
	vsum.y = vecta.y + vectb.y
	vsum.z = vecta.z + vectb.z
	
	return vsum
end

local function VectorBounce(mo,vmom,vslopenorm,percent)
	local player = mo.player
	if (player == nil) or (player.playerstate != PST_LIVE) return end
	
	local vbounce = Percent3D(VectAdd3D(FixedScalar3D(FixedScalar3D(vslopenorm, FixedDotProduct3D(vmom,vslopenorm)), -2 * FRACUNIT), vmom), percent)
	
	mo.momx = vbounce.x
	mo.momy = vbounce.y
	mo.momz = vbounce.z
end


local function DoWallBounce(mo,player,wallnormangle,walltype,side,reflect,slope)
	--Wall type
	local bouncy = (walltype == 2)
	local hbouncy = (walltype == 3)
	local vbouncy = (walltype == 4)
	local nocl = (walltype == 1)

	--Hold jump button for better bounce, don't hold for small bounce
	local bigbounce = player.holdingjump
	
	--Calculate angle
	local vwallnorm
	if slope == nil
		vwallnorm = SphereToCartesian(wallnormangle, 0)
	else
		vwallnorm = slope.normal
	end
	
	--Noclimb wall
	if nocl
		bigbounce = false
		S_StartSound(mo,sfx_s3k9e)
	end
	
	--Bounce stats based on the situation
	local percent = BOUNCEPERCENT
	if not bigbounce
		percent = WEAKBOUNCEPERCENT
	end
	local bouncez = WEAKWALLBOUNCEZ
	
	--Hold jump button for better bounce, don't hold for small bounce
	if bigbounce and not nocl
		bouncez = WALLBOUNCEZ
	elseif nocl
		bouncez = NOCLWALLBOUNCEZ
	end
	
	--Screenshake
	if player == consoleplayer and (bigbounce or nocl)
		local shake = 12
		local shaketics = 3
		if player.powers[pw_super]
			shake = 18
			shaketics = 5
		end
		P_StartQuake(shake * FRACUNIT, shaketics)
	end
	
	--mobj momentum vector
	local vmom = {}
	vmom.x = mo.momx
	vmom.y = mo.momy
	vmom.z = mo.momz
	
	--Reset jump flags
	player.pflags = $ | PF_JUMPED
	player.pflags = $ | PF_STARTJUMP
	
	local weak = false
	if not bigbounce and not nocl
		--Weak bump sfx
		S_StartSoundAtVolume(mo,sfx_s3k5d,155)
		weak = true
	end
	
	--Double dash is allowed after a wall jump
	if player.doubledash == nil
		player.doubledash = true
		--1st bounce sfx
		if not weak
			S_StartSoundAtVolume(mo,sfx_bounc1,175)
		end
	else
		player.doubledash = false
		mo.state = S_PLAY_SPRING
		bouncez = $ * SECONDBOUNCEZMULT / 100
		percent = $ + 12
		player.pflags = $ | PF_NOJUMPDAMAGE
		--2nd bounce sfx
		if not weak
			S_StartSound(mo,sfx_bounc2)
		end
	end
	
	--Wallthrust (the strength of momentum applied in the direction of the wall norm)
	local wallth = WALLTHRUST
	if not bigbounce
		wallth = WEAKWALLTHRUST
	end
	if nocl
		wallth = NOCLIMBWALLTHRUST
	end
	
	--Wall type
	if bouncy
		S_StartSound(mo,sfx_s3k87)
		bouncez = $ + 6*FRACUNIT
		percent = $ + 6
	end
	if hbouncy
		wallth = $ + 40*FRACUNIT
		S_StartSound(mo,sfx_cdfm74)
	end
	if vbouncy
		wallth = $ / 2
		percent = $ / 4
		bouncez = $ + 11*FRACUNIT
		S_StartSound(mo,sfx_cdfm62)
	end
	
	--Do the horizontal bounce
	if reflect == 1--wall
		VectorBounce(mo,vmom,vwallnorm,percent)
		
		if slope != nil--slope bounce
			mo.momz = wallth
			
		else--wall bounce
			if side == 1
				P_Thrust(mo, wallnormangle, wallth)
				player.drawangle = wallnormangle
			else
				P_Thrust(mo, wallnormangle, -wallth)
				player.drawangle = wallnormangle + ANGLE_180
			end
		end
	elseif reflect == 0--boss
		percent = $ - 10
		mo.momz = $ + 2 * FRACUNIT
		mo.momx = $ * percent / 100
		mo.momy = $ * percent / 100
		P_Thrust(mo, wallnormangle, wallth)
		player.drawangle = wallnormangle
	elseif reflect == 2--badnik/monitor
		local zm = max(0, (player.dashticsleft * 6/10) + 38)//The badnik bounce force reduces based on the dash length
		bouncez = max(abs(mo.momz * (zm) / 100), $) * BADNIKZMULT / 100
		if bigbounce
			mo.momx = $ * BADNIKSPEED / 100
			mo.momy = $ * BADNIKSPEED / 100
		else
			mo.momx = $ * BADNIKBUMP / 100
			mo.momy = $ * BADNIKBUMP / 100
		end
	elseif reflect == 3--egg colosseum
		mo.momz = $ * 3 / 4
		player.drawangle = wallnormangle
	end
	
	--Reset stuff
	player.bounceline = nil
	player.dashticsleft = nil
	
	--Vertical boost
	if mo.eflags & MFE_UNDERWATER
		bouncez = FixedMul($, WATERFACTOR)
	end
	if player.powers[pw_super]
		bouncez = FixedMul($, SUPERFACTOR)
	end
	if mo.eflags & MFE_VERTICALFLIP
		mo.momz = min($,-bouncez)
	else
		mo.momz = max($,bouncez)
	end
	
	--Wallbounce dust
	if nocl
		local dustcount = 6
		while dustcount > 0
			dustcount = $ - 1
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, mo.scale * 18, MT_SPINDUST)
			dust.state = S_MINECARTSPARK
			dust.momx = (mo.momx * 3) + (P_RandomRange(-20,20) * FRACUNIT)
			dust.momy = (mo.momy * 3) + (P_RandomRange(-20,20) * FRACUNIT)
			dust.momz = P_RandomRange(-3,3) * FRACUNIT
			dust.scale = FRACUNIT * P_RandomRange(50,125) / 100
			dust.destscale = 0
			dust.angle = P_RandomRange(0,ANGLE_180)
			dust.fuse = 10
		end
	else
		local dustcount = 10
		if not bigbounce
			dustcount = 5
		end
		while dustcount > 0
			dustcount = $ - 1
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, mo.scale * 18, MT_SPINDUST)
			dust.momx = (mo.momx * 2) + (P_RandomRange(-10,10) * FRACUNIT)
			dust.momy = (mo.momy * 2) + (P_RandomRange(-10,10) * FRACUNIT)
			dust.momz = P_RandomRange(-3,3) * FRACUNIT
			dust.scale = FRACUNIT * P_RandomRange(50,125) / 100
			dust.destscale = 0
			dust.color = mo.color
			dust.colorized = true
			dust.state = S_BOSSEXPLODE
		end
	end
	
	if bigbounce
		--Wallbounce big puff of smoke
		local boom = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_SPINDUST)
		boom.state = S_XPLD3
		boom.colorized = true
		boom.color = mo.color
	end
	
	--Wallbounce circle effect
	local circle = P_SpawnMobjFromMobj(mo, 0, 0, mo.scale * 19, MT_REBOUND)
	circle.angle = wallnormangle + ANGLE_90
	circle.fuse = 9
	circle.scale = 0
	circle.destscale = 10*FRACUNIT
	circle.color = mo.color
	circle.colorized = true
	
	if not bigbounce
		circle.fuse = 6
	end
end


--Rebound dash
addHook("AbilitySpecial", function(player)
	local mo = player.mo
	if not (mo and mo.valid and mo.skin == rbdskin) return false end
	if (player.pflags & (PF_THOKKED|PF_SHIELDABILITY)) and player.doubledash != true return true end
	
	if player.doubledash == true
		player.doubledash = false
	end
	
	--Flags, timers, state
	mo.state = S_PLAY_JUMP
	player.pflags = $1 | PF_JUMPED
	player.pflags = $1 | PF_THOKKED
	player.pflags = $1 & ~(PF_SPINNING|PF_STARTDASH)
	player.dashticsleft = DASHLENGTH * 10
	player.holdingjump = true
	
	--Screenshake
	if player == consoleplayer and player.powers[pw_super]
		local shake = 10
		local shaketics = 3
		P_StartQuake(shake * FRACUNIT, shaketics)
	end
	
	--XY Momentum, timer
	local dashsp = player.actionspd
	if mo.eflags & MFE_UNDERWATER
		dashsp = FixedMul($, WATERFACTOR)
	end
	if player.powers[pw_super]
		dashsp = FixedMul($, SUPERFACTOR)
	end
	
	local momangle = R_PointToAngle2(0,0,player.rmomx,player.rmomy)
	local anglediff = AngleFixed(abs(momangle - mo.angle)) / FRACUNIT
	local pspd = player.speed
	
	local speedpercentage = max(0, min(100, (100 - (anglediff - 50))))
	pspd = $ * speedpercentage / 100
	
	P_InstaThrust(mo, mo.angle, FixedMul(max(dashsp,pspd), mo.scale))
	
	//Z momentum
	if mo.eflags & MFE_VERTICALFLIP
		if mo.momz > 0
			mo.momz = $ * DASHZPERCENT / 100
		end
	else
		if mo.momz < 0
			mo.momz = $ * DASHZPERCENT / 100
		end
	end
	
	--Visual and sfx
	player.lockangle = mo.angle
	S_StartSound(mo,sfx_airdsh)
	local circle = P_SpawnMobjFromMobj(mo, 0, 0, mo.scale * 24, MT_REBOUND)
	circle.angle = mo.angle + ANGLE_90
	circle.fuse = 7
	circle.scale = FRACUNIT / 3
	circle.destscale = 10*FRACUNIT
	circle.colorized = true
	circle.color = mo.color
	circle.momx = -mo.momx / 2
	circle.momy = -mo.momy / 2
	
	return true
end)

addHook("ThinkFrame", function()
	for player in players.iterate()
		local mo = player.mo
		if not (mo and mo.valid and mo.skin == rbdskin) continue end
		if (player == nil) or (player.playerstate != PST_LIVE) continue end
		
		if player.prevpflags != nil
			--Roll after dash
			if P_IsObjectOnGround(mo) and (player.prevpflags & PF_THOKKED) and (player.cmd.buttons & BT_USE)
			and (FixedHypot(mo.momx, mo.momy) > (5 * mo.scale))
				mo.state = S_PLAY_ROLL
				player.pflags = $ | PF_SPINNING
				S_StartSound(mo,sfx_spin)
			end
		end
		
		--Reset dash counter and timer
		if (P_IsObjectOnGround(mo) or not (player.pflags & PF_THOKKED))
			player.doubledash = nil
			player.dashticsleft = nil
			player.holdingjump = nil
		end
		
		player.prevpflags = player.pflags
	end
end)

local framelist = {0, 1, 3, 4}
addHook("PostThinkFrame", function()
	for player in players.iterate()
		local mo = player.mo
		if not (mo and mo.valid and mo.skin == rbdskin) continue end
		if (player == nil) or (player.playerstate != PST_LIVE) continue end
		
		if (mo.previousx == nil) or (mo.previousy == nil) or (mo.previousz == nil)
			mo.previousx = mo.x
			mo.previousy = mo.y
			mo.previousz = mo.z
		end
		
		if (player.speed > 15*FRACUNIT) and (player.pflags & PF_SPINNING) and not (player.pflags & PF_JUMPED)
			for i = 0, 9
				local percent = FRACUNIT * (i * 10) / 100
				local trail = P_SpawnGhostMobj(mo)
				local tx = FixedLerp(mo.x,mo.previousx,percent)
				local ty = FixedLerp(mo.y,mo.previousy,percent)
				local tz = FixedLerp(mo.z + 3*FRACUNIT,mo.previousz + 4*FRACUNIT,percent)
				P_TeleportMove(trail, tx, ty, tz - 7 * FRACUNIT)
				trail.fuse = 13
				trail.state = S_THOK
				trail.scalespeed = FRACUNIT/12
				trail.scale = (FRACUNIT * 5/6) - (i * FRACUNIT/120)
				trail.destscale = 0
				trail.frame = TR_TRANS70|A
			end
		end
		
		if player.doubledash and  mo.state == S_PLAY_JUMP
			mo.frame = framelist[(leveltime % 4) + 1]
			local g = P_SpawnGhostMobj(mo)
			g.frame = 2 | TR_TRANS70
			g.tics = 1
			
			if (leveltime % 3 == 0)
				S_StartSoundAtVolume(mo, sfx_s3k82, 40)
				g = P_SpawnGhostMobj(mo)
				g.frame = 2 | TR_TRANS90
				g.tics = 5
				g.destscale = 0
			end
		end
		
		if player.dashticsleft
			player.drawangle = player.lockangle
			
			if mo.state != S_PLAY_JUMP
				player.doubledash = nil
				player.dashticsleft = nil
				
			else
				--Check if the player is still holding jump
				if player.holdingjump
					player.holdingjump = (player.cmd.buttons & BT_JUMP)
					local holdingforward
					local fm = player.cmd.forwardmove
					local sm = player.cmd.sidemove
					
					if player.pflags&PF_ANALOGMODE --god damn it fuck
						local angleturn = player.cmd.angleturn<<FRACBITS
						local angdiff = player.lockangle - angleturn
						local sine = sin(angdiff)
						local cosine = cos(angdiff)
						local realforward = fm*cosine - sm*sine
						
						holdingforward = realforward >= 10*FRACUNIT
					else
						holdingforward = (fm >= 10)
					end
					
					--Short dash by letting go of the button and not holding forward
					if not player.holdingjump and not holdingforward
						if player.dashticsleft > 10*DASHLENGTH - 30
							mo.momx = FixedDiv($, 2 * FRACUNIT)
							mo.momy = FixedDiv($, 2 * FRACUNIT)
						elseif player.dashticsleft > 10*DASHLENGTH - 50
							mo.momx = FixedDiv(FixedMul($, 2 * FRACUNIT), 3 * FRACUNIT)
							mo.momy = FixedDiv(FixedMul($, 2 * FRACUNIT), 3 * FRACUNIT)
						elseif player.dashticsleft > 10*DASHLENGTH - 60
							mo.momx = FixedDiv(FixedMul($, 3 * FRACUNIT), 4 * FRACUNIT)
							mo.momy = FixedDiv(FixedMul($, 3 * FRACUNIT), 4 * FRACUNIT)
						elseif player.dashticsleft > 10*DASHLENGTH - 70
							mo.momx = FixedDiv(FixedMul($, 4 * FRACUNIT), 5 * FRACUNIT)
							mo.momy = FixedDiv(FixedMul($, 4 * FRACUNIT), 5 * FRACUNIT)
						end
					end
				end
				if player.holdingjump
					--Thok trail (which disappears a bit earlier in order to make sure the player *really* knows they messed up the timing when they hit a wall too late)
					if player.dashticsleft > 80
						for i = 0, 9
							local percent = FRACUNIT * (i * 10) / 100
							local trail = P_SpawnGhostMobj(mo)
							local tx = FixedLerp(mo.x,mo.previousx,percent)
							local ty = FixedLerp(mo.y,mo.previousy,percent)
							local tz = FixedLerp(mo.z + 3*FRACUNIT,mo.previousz + 4*FRACUNIT,percent)
							P_TeleportMove(trail, tx, ty, tz - 7 * FRACUNIT)
							trail.fuse = 13
							trail.state = S_THOK
							trail.scalespeed = FRACUNIT/12
							trail.scale = (FRACUNIT * 4/5) - (i * FRACUNIT/120)
							trail.destscale = 0
							trail.frame = TR_TRANS70|A
						end
					end
					player.dashticsleft = $ - 10
				else
					--Lesser thok trail (player let go of the jump button, so they can't wallbounce anymore)
					if player.dashticsleft > 80
						for i = 0, 9
							local percent = FRACUNIT * (i * 10) / 100
							local trail = P_SpawnGhostMobj(mo)
							local tx = FixedLerp(mo.x,mo.previousx,percent)
							local ty = FixedLerp(mo.y,mo.previousy,percent)
							local tz = FixedLerp(mo.z + 3*FRACUNIT,mo.previousz + 4*FRACUNIT,percent)
							P_TeleportMove(trail, tx, ty, tz - 3 * FRACUNIT)
							trail.fuse = 9
							trail.state = S_THOK
							trail.scalespeed = FRACUNIT/15
							trail.scale = (FRACUNIT * 3/5) - (i * FRACUNIT/150)
							trail.destscale = 0
							trail.frame = TR_TRANS80|A
						end
					end
					player.dashticsleft = $ - 15
				end
			end
		end
		
		--Reset bounceline
		player.bounceline = nil
		mo.previousx = mo.x
		mo.previousy = mo.y
		mo.previousz = mo.z
	end
end)

addHook("MobjLineCollide", function(mo, line)
	if not mo or not mo.valid return end
	if not (mo.skin == rbdskin) return end
	local player = mo.player
	if (player == nil) or (player.playerstate != PST_LIVE) return end
	if (not player.dashticsleft) or (player.dashticsleft <= 0) return end
	
	--Horizon line
	if line.special == 41 return end
	
	--Set the bounceline so we can bounce off it later in case that line actually blocked the player from moving
	local side = P_PointOnLineSide(mo.x,mo.y,line)
	local sector = nil
	
	--One-sided walls
	if line.backsector == nil
		player.bounceline = line
		player.bounceside = 0
		return
	end
	if line.frontsector == nil
		player.bounceline = line
		player.bounceside = 1
		return
	end
	
	--Which side are we hitting the wall from?
	if side == 1
		sector = line.frontsector
	else
		sector = line.backsector
	end
	
	--Impassible line
	if (line.flags & ML_IMPASSIBLE) and line.frontside.midtexture
		player.bounceline = line
		player.bounceside = side
		return
	end
	
	if sector == nil return end
	
	--Polyobject
	for i = 0, #sector.lines
		local li = sector.lines[i]
		if li == nil continue end
		
		if li.special == 20--First line of polyobject
			local topheight = sector.ceilingheight
			local bottomheight = sector.floorheight
			
			if (topheight < mo.z)
				return
			end

			if (bottomheight > mo.z + mo.height)
				return
			end

			player.bounceline = line
			player.bounceside = side
			return
		end
	end
	
	--Standard
	local ceilz = sector.ceilingheight
	if sector.c_slope
		ceilz = P_GetZAt(sector.c_slope, mo.x, mo.y)
	end
	if (ceilz < mo.z + mo.height)
		if sector.ceilingpic != "F_SKY1"
			player.bounceline = line
			player.bounceside = side
		end
		return
	end
	local floorz = sector.floorheight
	if sector.f_slope
		floorz = P_GetZAt(sector.f_slope, mo.x, mo.y)
	end
	if (floorz > mo.z)
		player.bounceline = line
		player.bounceside = side
		return
	end
	
	--FOF
	for rover in sector.ffloors()
		if not (rover.flags & FF_EXISTS) or not (rover.flags & FF_BLOCKPLAYER)
			continue
		end

		local topheight = rover.topheight
		local bottomheight = rover.bottomheight

		if (rover.t_slope)
			topheight = P_GetZAt(rover.t_slope, mo.x, mo.y)
		end
		if (rover.b_slope)
			bottomheight = P_GetZAt(rover.b_slope, mo.x, mo.y)
		end

		if (topheight < mo.z)
			continue
		end

		if (bottomheight > mo.z + mo.height)
			continue
		end
	
		player.bounceline = line
		player.bounceside = side
		if (rover.flags & FF_BUSTUP) and not (rover.flags & FF_STRONGBUST)
			player.bustsector = sector
			player.bustrover = rover
		end
		break
	end
	
	--Solid midtexture
	--TODO: once textures[] is exposed to lua, change this so it properly checks the height of a solid midtexture and decides if it's blocking the player or not.
	if (line.flags & ML_EFFECT4) and line.frontside.midtexture
		player.bounceline = line
		player.bounceside = side
		return
	end
end, MT_PLAYER)


addHook("PreThinkFrame", function()	
	for player in players.iterate()
		local mo = player.mo
		if not (mo and mo.valid and mo.skin == rbdskin) return end
		if (player == nil) or (player.playerstate != PST_LIVE) return end
	end
end)

addHook("MobjDeath", function(enemy, mo)
	if not mo or not mo.valid return end
	if not (mo.skin == rbdskin) return end
	local player = mo.player
	if (player == nil) or (player.playerstate != PST_LIVE) return end
	if (not player.dashticsleft) or (player.dashticsleft <= 0) return end
	
	if enemy and enemy.valid and (enemy.flags & MF_MONITOR)
		local wallnormangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
		DoWallBounce(mo,player,wallnormangle,false,0,2)
	end
end)

addHook("MobjDamage", function(enemy, mo)
	if not mo or not mo.valid return end
	if not (mo.skin == rbdskin) return end
	local player = mo.player
	if (player == nil) or (player.playerstate != PST_LIVE) return end
	if (not player.dashticsleft) or (player.dashticsleft <= 0) return end
	
	if enemy and enemy.valid and (enemy.flags & MF_ENEMY or enemy.flags & MF_BOSS)
		local wallnormangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
		if enemy.type == MT_EGGMOBILE4
			DoWallBounce(mo,player,wallnormangle,false,0,3)
		elseif (enemy.info.spawnhealth > 1)
			DoWallBounce(mo,player,wallnormangle,false,0,0)
		else
			DoWallBounce(mo,player,wallnormangle,false,0,2)
		end
	end
end)

addHook("MobjMoveCollide", function(mo, other)
	if not mo or not mo.valid return end
	if not (mo.skin == rbdskin) return end
	local player = mo.player
	if (player == nil) or (player.playerstate != PST_LIVE) return end
	if (not player.dashticsleft) or (player.dashticsleft <= 0) return end
	if (other.flags & MF_SOLID) and not (other.flags & MF_ENEMY or other.flags & MF_BOSS or other.flags & MF_MONITOR or other.flags & MF_PAIN) and not (other.type == MT_PLAYER)
		if (mo.z < other.z + other.height) and (mo.z + mo.height >= other.z)
			if (other.flags & MF_PUSHABLE)
				other.momx = $ + (mo.momx / 2)
				other.momy = $ + (mo.momy / 2)
				other.momz = 3 * FRACUNIT
			end
			local wallnormangle = R_PointToAngle2(mo.x - mo.momx, mo.y - mo.momy, other.x, other.y)
			DoWallBounce(mo,player,wallnormangle,false,0,1)
			return true
		end
	end
end, MT_PLAYER)

addHook("MobjMoveBlocked", function(mo)//, thing, line)
	local player = mo.player
	if (not player.dashticsleft) or (player.dashticsleft <= 0) return end
	
	/*if thing and not (thing.flags & MF_ENEMY or thing.flags & MF_BOSS or thing.flags & MF_MONITOR or thing.flags & MF_PAIN or thing.type == MT_PLAYER)
		if (thing.flags & MF_PUSHABLE)
			thing.momx = $ + (mo.momx / 2)
			thing.momy = $ + (mo.momy / 2)
			thing.momz = 3 * FRACUNIT
		end
		local wallnormangle = R_PointToAngle2(mo.x - mo.momx, mo.y - mo.momy, thing.x, thing.y)
		DoWallBounce(mo,player,wallnormangle,false,0,1)
		return true
	end*/
	
	local line = player.bounceline
	if line == nil return end
	local side = player.bounceside//P_PointOnLineSide(mo.x - mo.momx, mo.y - mo.momy, line)
	if side == nil return end

	--Noclimb walls result in a super weak bounce
	local nocl = line.flags & ML_NOCLIMB
	
	--Bouncy walls result in a beeg bounce
	local bouncy = (line.special == 999)
	local hbouncy = (line.special == 998)
	local vbouncy = (line.special == 997)
	
	--Bustable FOFs are busted
	if player.bustsector and player.bustrover and player.holdingjump
		EV_CrumbleChain(player.bustsector, player.bustrover)
		player.bustsector = nil
		player.bustrover = nil
	end
	
	local wallnormangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y) + ANGLE_90
	
	if bouncy
		DoWallBounce(mo,player,wallnormangle,2,side,1)
	elseif hbouncy
		DoWallBounce(mo,player,wallnormangle,3,side,1)
	elseif vbouncy
		DoWallBounce(mo,player,wallnormangle,4,side,1)
	else
		if nocl
			nocl = 1
		end
		DoWallBounce(mo,player,wallnormangle,nocl,side,1)
	end
end, MT_PLAYER)